#include <stdio.h>
#include <stdlib.h>
#include <math.h>
#include <errno.h>
#include <assert.h>

#include "el.h"
#include "fpe.h"

/* Call is_error when errno != 0, and where x is the result libm
 * returned.  is_error will usually set up an exception and return
 * true (1), but may return false (0) without setting up an exception.
 */
static int
is_error (RvObject *self, double x)
{
  int result = 1;	/* presumption of guilt */
  assert (errno);	/* non-zero errno is a precondition for calling */
  if (errno == EDOM)
    RvLogError (self, ERRLOG_ERR, ERRLOG_MARK, "math domain error");
  else if (errno == ERANGE) 
    {
      /* ANSI C generally requires libm functions to set ERANGE
       * on overflow, but also generally *allows* them to set
       * ERANGE on underflow too.  There's no consistency about
       * the latter across platforms.
       * Alas, C99 never requires that errno be set.
       * Here we suppress the underflow errors (libm functions
       * should return a zero on underflow, and +- HUGE_VAL on
       * overflow, so testing the result for zero suffices to
       * distinguish the cases).
       */
      if (x)
        RvLogError (self, ERRLOG_ERR, ERRLOG_MARK, "math range error");
      else
        result = 0;
    }
  else
    /* Unexpected math error */
    RvLogError (self, ERRLOG_ERR, ERRLOG_MARK, "Unexpected math error");

  return result;
}

static int
math_1 (self, args, ret, func)
     RvObject *self;
     RvValue *args, *ret;
     double (*func) (double);
{
  double x;
  
  if (RvParseArg (self, args, "d", &x))
    return -1;
  
  errno = 0;
  RvFPE_START_PROTECT ("in math_1", return -1);
  x = (*func) (x);
  RvFPE_END_PROTECT (x);

  if (errno == 0)
    {
      if ((x) == HUGE_VAL || (x) == -HUGE_VAL)
        errno = ERANGE;
      else if (isnan (x)) 
        errno = EDOM;
    }
  if (errno && is_error (self, x))
    return -1;

  RvSetReturn (self, ret, "D", x);

  return 0;
}

static int
math_2 (self, args, ret, func)
     RvObject *self;
     RvValue *args, *ret;
     double (*func) (double, double);
{
  double x, y;
  
  if (RvParseArg (self, args, "dd", &x, &y))
    return -1;
  
  errno = 0;
  RvFPE_START_PROTECT ("in math_1", return -1);
  x = (*func) (x, y);
  RvFPE_END_PROTECT (x);

  if (errno == 0)
    {
      if ((x) == HUGE_VAL || (x) == -HUGE_VAL)
        errno = ERANGE;
      else if (isnan (x)) 
        errno = EDOM;
    }
  if (errno && is_error (self, x))
    return -1;

  RvSetReturn (self, ret, "D", x);

  return 0;
}

#define FUNC1(funcname)                         \
  static int                                    \
  math_##funcname##_td (self, args, ret)        \
       RvObject *self;                          \
       RvValue *args, *ret;                     \
  {                                             \
    return math_1 (self, args, ret, funcname);  \
  }

#define FUNC2(funcname)                         \
  static int                                    \
  math_##funcname##_tdd (self, args, ret)       \
       RvObject *self;                          \
       RvValue *args, *ret;                     \
  {                                             \
    return math_2 (self, args, ret, funcname);  \
  }

FUNC1 (acos)
FUNC1 (asin)
FUNC1 (atan)
FUNC2 (atan2)
FUNC1 (ceil)
FUNC1 (cos)
FUNC1 (cosh)
FUNC1 (exp)
FUNC1 (fabs)
FUNC1 (floor)
FUNC2 (fmod)
FUNC2 (hypot)
FUNC2 (pow)
FUNC1 (sin)
FUNC1 (sinh)
FUNC1 (sqrt)
FUNC1 (tan)
FUNC1 (tanh)

static const double degToRad = 3.141592653589793238462643383 / 180.0;

static int
math_degrees_td (self, args, ret)
     RvObject *self;
     RvValue *args, *ret;
{
  double x;
  
  if (RvParseArg (self, args, "d", &x))
    return -1;

  RvSetReturn (self, ret, "D", x / degToRad);

  return 0;
}

static int
math_radians_td (self, args, ret)
     RvObject *self;
     RvValue *args, *ret;
{
  double x;
  
  if (RvParseArg (self, args, "d", &x))
    return -1;

  RvSetReturn (self, ret, "D", x * degToRad);

  return 0;
}

static int
math_frexp_td (self, args, ret)
     RvObject *self;
     RvValue *args, *ret;
{
  double x;
  int i;
  RvObject *result;
  
  if (RvParseArg (self, args, "d", &x))
    return -1;
  errno = 0;
  x = frexp (x, &i);
  
  if (errno == 0)
    {
      if ((x) == HUGE_VAL || (x) == -HUGE_VAL)
        errno = ERANGE;
      else if (isnan (x)) 
        errno = EDOM;
    }
  if (errno && is_error (self, x))
    return -1;

  result = RvArrayNew (self, 0);
  RvArrayPush (result, "D", x);
  RvArrayPush (result, "I", i);
  RvSetReturn (self, ret, "[", result);

  return 0;
}

static int
math_ldexp_tdi (self, args, ret)
     RvObject *self;
     RvValue *args, *ret;
{
  double x;
  int exp;
  
  if (RvParseArg (self, args, "di", &x, &exp))
    return -1;

  errno = 0;
  RvFPE_START_PROTECT ("ldexp", return -1);
  x = ldexp (x, exp);
  RvFPE_END_PROTECT (x);

  if (errno == 0)
    {
      if ((x) == HUGE_VAL || (x) == -HUGE_VAL)
        errno = ERANGE;
      else if (isnan (x)) 
        errno = EDOM;
    }
  if (errno && is_error (self, x))
    return -1;

  RvSetReturn (self, ret, "D", x);

  return 0;
}

static int
math_log_td (self, args, ret)
     RvObject *self;
     RvValue *args, *ret;
{
  return math_1 (self, args, ret, log);
}

static int
math_log10_td (self, args, ret)
     RvObject *self;
     RvValue *args, *ret;
{
  return math_1 (self, args, ret, log10);
}

static int
math_modf_td (self, args, ret)
     RvObject *self;
     RvValue *args, *ret;
{
  double x, y;
  RvObject *result;
  
  if (RvParseArg (self, args, "d", &x))
    return -1;
  errno = 0;
  x = modf (x, &y);
  
  if (errno == 0)
    {
      if ((x) == HUGE_VAL || (x) == -HUGE_VAL)
        errno = ERANGE;
      else if (isnan (x)) 
        errno = EDOM;
    }
  if (errno && is_error (self, x))
    return -1;

  result = RvArrayNew (self, 0);
  RvArrayPush (result, "D", x);
  RvArrayPush (result, "D", y);
  RvSetReturn (self, ret, "[", result);

  return 0;
}

static RvMethodDef method[] =
  {
    { "acos",           "(TD)D",        math_acos_td },
    { "asin",           "(TD)D",        math_asin_td },
    { "atan",           "(TD)D",        math_atan_td },
    { "atan2",          "(TDD)D",       math_atan2_tdd },
    { "ceil",           "(TD)D",        math_ceil_td },
    { "cos",            "(TD)D",        math_cos_td },
    { "cosh",           "(TD)D",        math_cosh_td },
    { "degrees",        "(TD)D",        math_degrees_td },
    { "exp",            "(TD)D",        math_exp_td },
    { "fabs",           "(TD)D",        math_fabs_td },
    { "floor",          "(TD)D",        math_floor_td },
    { "fmod",           "(TDD)D",       math_fmod_tdd },
    { "frexp",          "(TD)[",        math_frexp_td },
    { "hypot",          "(TDD)D",       math_hypot_tdd },
    { "ldexp",          "(TDI)D",       math_ldexp_tdi },
    { "log",            "(TD)D",        math_log_td },
    { "log10",          "(TD)D",        math_log10_td },
    { "modf",           "(TD)[",        math_modf_td },
    { "pow",            "(TDD)D",       math_pow_tdd },
    { "radians",        "(TD)D",        math_radians_td },
    { "sin",            "(TD)D",        math_sin_td },
    { "sinh",           "(TD)D",        math_sinh_td },
    { "sqrt",           "(TD)D",        math_sqrt_td },
    { "tan",            "(TD)D",        math_tan_td },
    { "tanh",           "(TD)D",        math_tanh_td },
  };

/** entry point.  */
int
init_math (RvClass *cls)
{
  if (RvDefineMethods (cls, method))
    return -1;

  return 0;
}
